"""
Sound of Pixels code taken from:
https://github.com/hangzhaomit/Sound-of-Pixels/tree/master/models
"""

import torch
import torchvision
import torch.nn.functional as F

from .synthesizer_net import InnerProd, Bias
from .audio_net import Unet
from .vision_net import ResnetFC, ResnetDilated
from .criterion import BCELoss, L1Loss, L2Loss

__all__ = ['get_network']


def activate(x, activation):
    if activation == 'sigmoid':
        return torch.sigmoid(x)
    elif activation == 'softmax':
        return F.softmax(x, dim=1)
    elif activation == 'relu':
        return F.relu(x)
    elif activation == 'tanh':
        return F.tanh(x)
    elif activation == 'no':
        return x
    else:
        raise Exception('Unkown activation!')


class ModelBuilder():
    # custom weights initialization
    def weights_init(self, m):
        classname = m.__class__.__name__
        if classname.find('Conv') != -1:
            m.weight.data.normal_(0.0, 0.001)
        elif classname.find('BatchNorm') != -1:
            m.weight.data.normal_(1.0, 0.02)
            m.bias.data.fill_(0)
        elif classname.find('Linear') != -1:
            m.weight.data.normal_(0.0, 0.0001)

    def build_sound(self, arch='unet5', fc_dim=64, weights=''):
        # 2D models
        if arch == 'unet5':
            net_sound = Unet(fc_dim=fc_dim, num_downs=5)
        elif arch == 'unet6':
            net_sound = Unet(fc_dim=fc_dim, num_downs=6)
        elif arch == 'unet7':
            net_sound = Unet(fc_dim=fc_dim, num_downs=7)
        else:
            raise Exception('Architecture undefined!')

        net_sound.apply(self.weights_init)
        if len(weights) > 0:
            print('Loading weights for net_sound')
            net_sound.load_state_dict(torch.load(weights))

        return net_sound

    # builder for vision
    def build_frame(self, arch='resnet18', fc_dim=64, pool_type='avgpool',
                    weights=''):
        pretrained = True
        if arch == 'resnet18fc':
            original_resnet = torchvision.models.resnet18(pretrained)
            net = ResnetFC(
                original_resnet, fc_dim=fc_dim, pool_type=pool_type)
        elif arch == 'resnet18dilated':
            original_resnet = torchvision.models.resnet18(pretrained)
            net = ResnetDilated(
                original_resnet, fc_dim=fc_dim, pool_type=pool_type)
        else:
            raise Exception('Architecture undefined!')

        if len(weights) > 0:
            print('Loading weights for net_frame')
            net.load_state_dict(torch.load(weights))
        return net

    def build_synthesizer(self, arch, fc_dim=64, weights=''):
        if arch == 'linear':
            net = InnerProd(fc_dim=fc_dim)
        elif arch == 'bias':
            net = Bias()
        else:
            raise Exception('Architecture undefined!')

        net.apply(self.weights_init)
        if len(weights) > 0:
            print('Loading weights for net_synthesizer')
            net.load_state_dict(torch.load(weights))
        return net


class NetWrapper(torch.nn.Module):
    def __init__(self, nets):
        super(NetWrapper, self).__init__()
        self.net_sound, self.net_frame, self.net_synthesizer = nets

    def forward(self,*args):
        log_mag_mix, frames=args

        N = 2
        # B = mag_mix.size(0)
        # T = mag_mix.size(3)


        # 1. forward net_sound -> BxCxHxW
        feat_sound = self.net_sound(log_mag_mix)

        # 2. forward net_frame -> Bx1xC
        feat_frames= self.net_frame.forward_multiframe(frames)
        feat_frames= activate(feat_frames, 'sigmoid')

        # 3. sound synthesizer
        pred_masks = self.net_synthesizer(feat_frames, feat_sound)
        # pred_masks = activate(pred_masks, 'sigmoid')

        # 4. loss

        return pred_masks


def get_network():
    builder = ModelBuilder()
    net_sound = builder.build_sound(
        arch="unet7",
        fc_dim=32)
    net_frame = builder.build_frame(
        arch='resnet18dilated',
        fc_dim=32,
        pool_type='maxpool')
    net_synthesizer = builder.build_synthesizer(
        arch='linear',
        fc_dim=32)
    nets = (net_sound, net_frame, net_synthesizer)
    return NetWrapper(nets)
